import numpy as np
import warnings
from .KMM import KernelMeanMatching


class KMMTransferReg():
    """
    KMMTransferReg:
    KMMTransferReg is a transfer learning regression model based on Kernel Mean Matching (KMM) algorithm. It maps
    the source and target datasets to a common feature space, where the marginal distributions of the source and target
    data are as close as possible, while maintaining the conditional distribution of the source data. The mapped data
    is then used to train a regression model that can make predictions on the target dataset.

    Author:
    ----------
    Bin Cao, ZheJiang LAB, Hangzhou, CHINA.

    Parameters
    ----------
    Regressor : str or object, default='RF'
        A regression model used to fit the mapped data. If a string is passed, it should be one of {'RF', 'LR'}
        representing RandomForestRegressor and LinearRegression, respectively. Otherwise, an object with fit and
        predict methods that implements a regression algorithm can be passed.
    UpBound : float, default=1
        The upper bound of the beta for the target dataset during kernel mean matching.
    kernel : str, default='RBF'
        Kernel function used for kernel mean matching. It should be one of {'RBF', 'DotProduct', 'WhiteKernel', 'Matern'}
        representing the Radial Basis Function kernel, Dot Product kernel, White noise kernel and Matern kernel,
        respectively.
    Targets : int, default=1
        The number of target variables to be predicted.

    Methods
    -------
    fit(source_dataset,target_dataset,test_data)
        Fit the transfer learning regression model to the source dataset and target dataset, and predict the target variable
        on the test dataset.
    """
    def __init__(self, Regressor='RF',UpBound=1,kernel = 'RBF', Targets = 1):
        """
        Initialize KMMTransferReg model.

        Parameters
        ----------
        Regressor : str or object, default='RF'
            A regression model used to fit the mapped data. If a string is passed, it should be one of {'RF', 'LR'}
            representing RandomForestRegressor and LinearRegression, respectively. Otherwise, an object with fit and
            predict methods that implements a regression algorithm can be passed.
        UpBound : float, default=1
            The upper bound for beta coefficients. 
        kernel : str, default='RBF'
            The kernel to use for KMM. Can be 'RBF', 'DotProduct', 'WhiteKernel', 'Matern'.
        Targets : int, default=1
            The number of target variables in the dataset.
        """
        # A sklearn regression model 
        if type(Regressor) == str:
            self.Regressor = GenerateReg(Regressor)
        else:
            self.Regressor = Regressor
        self.UpBound = UpBound
        self.kernel = kernel
        self.Targets = Targets
        warnings.filterwarnings('ignore')


    def fit(self,source_dataset,target_dataset,test_data,tao=None):
        """
        Fit the transfer model on source and target datasets and return the predictions
        on the test data along with beta coefficients.

        Parameters
        ----------
        source_dataset : array-like of shape (n_samples, n_features)
            The source dataset, where n_samples is the number of samples and n_features
            is the number of features including the target variable(s).
        target_dataset : array-like of shape (n_samples, n_features)
            The target dataset, where n_samples is the number of samples and n_features
            is the number of features including the target variable(s).
        test_data : array-like of shape (n_samples, n_features - Targets)
            The test dataset, where n_samples is the number of samples and n_features
            is the number of features excluding the target variable(s).
        tao : float, default=None
            used in KMM : 1-tao <= sum(beta_i) <= 1+tao , E(beta) = 1 , E is expectation 
            if tao == None, tao =  B/np.sqrt(n), n is the number of source domain data.
        Returns
        -------
        predictions : array-like of shape (n_samples,)
            The predicted values for the test data.
        beta : array-like of shape (n_samples,)
            The beta coefficients derived by the KMM weighting scheme.
        """
        source_data = np.array(source_dataset)[:, :-self.Targets]
        source_response = np.array(source_dataset)[:, -self.Targets:]
        target_data = np.array(target_dataset)[:, :-self.Targets]
        target_response = np.array(target_dataset)[:, -self.Targets:]

        KMM = KernelMeanMatching(self.kernel,source_data,target_data,target_response)
        beta = KMM.cal_beta(B = self.UpBound,tao=tao)

        X = np.concatenate((source_data, target_data), axis=0)
        Y = np.concatenate((source_response, target_response), axis=0)

        # test the Regressor
        attribute_list = ['fit', 'predict',]
        check_attributes(self.Regressor, attribute_list)

        check_weight(self.Regressor.fit)

        data_weight = np.concatenate(
            (beta,np.ones(
                (len(target_data),1)
                        )))


        return self.Regressor.fit(X,Y, sample_weight = data_weight.flatten()).predict(test_data), beta.flatten()


def check_attributes(estimator, attribute_list):
    for attribute in attribute_list:
        if hasattr(estimator, attribute) :
            print(f"The estimator has a {attribute} attribute.")
            pass
        else:
            print(f"The estimator does not have a {attribute} attribute.")
            print('Please provide another Regressor')
            raise ValueError("Error of Regressor")


def check_weight(estimator,):
        if 'sample_weight' in estimator.__code__.co_varnames:
            print("The estimator.fit has 'sample_weight' attribute.")
            pass
        else:
            print("The estimator.fit does not have 'sample_weight' attribute.")
            print('Please provide another Regressor')
            raise ValueError("Error of Regressor")

def GenerateReg(Regressor_name):
    if Regressor_name == 'RF':
        from sklearn.ensemble import RandomForestRegressor
        mdoel = RandomForestRegressor
    elif Regressor_name == 'LR':
        from sklearn import linear_model
        mdoel = linear_model.LinearRegression()
    else:
        print('Sorry, Bin did not define this function for you, please pass it in yourself')
    return mdoel





print('KMMTransferReg : a transfer learning regression model based on Kernel Mean Matching (KMM) algorithm \n Bin CAO, ZheJiang LAB, Hangzhou, CHINA. (MGI, SHU, Shanghai, CHINA).')

__version__ = '1.0'
__description__ = 'a transfer learning regression model'
__author__ = 'Bin CAO'
__author_email__ = 'binjacobcao@gmail.com'
__url__ = 'https://github.com/Bin-Cao/KMMTransferRegressor'



import numpy as np
import copy
import warnings
from cvxopt import solvers, matrix 
from sklearn.gaussian_process import GaussianProcessRegressor as GPR
from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel, RBF, Matern 


class KernelMeanMatching():
    def __init__(self,kernel,source_data,target_data,target_response):
        """
        Constructor for KernelMeanMatching class.
        
        Args:
            kernel (str): Name of the kernel function to use.
            source_data (numpy array): n x d numpy array containing the source data features.
            target_data (numpy array): m x d numpy array containing the target data features.
            target_response (numpy array): m x 1(t) numpy array containing the target response values.
        """
        self.kernel = kernel
        # n data points
        self.source_data = source_data
        # m data points
        self.target_data = target_data
        self.target_response = target_response
        warnings.filterwarnings('ignore')

    def call_kernel(self,para = None):
        """
        Helper function to call the kernel function specified by the user.
        
        Args:
            para (numpy array): Parameters to use when instantiating the kernel function. Defaults to None.
        
        Returns:
            A kernel function object based on the user-specified kernel function and parameters.
        """
        if para is None :
            if self.kernel == 'RBF':
                kernel = RBF(length_scale_bounds=(0.01,100))  
            elif self.kernel == 'DotProduct':
                kernel = DotProduct(sigma_0_bounds=(0.01,100))
            elif self.kernel == 'WhiteKernel':
                kernel = WhiteKernel(noise_level_bounds=(0.01,100))
            elif self.kernel == 'Matern':
                kernel = Matern(length_scale_bounds=(0.01,100))
            else:
                print("Unknown kernel !")
                print('Only the following kernel functions are legal')
                print('RBF | DotProduct | WhiteKernel | Matern')
        else :
            if self.kernel == 'RBF':
                kernel = RBF(para,"fixed")
            elif self.kernel == 'DotProduct':
                kernel = DotProduct(para,"fixed")
            elif self.kernel == 'WhiteKernel':
                kernel = WhiteKernel(para,"fixed")
            elif self.kernel == 'Matern':
                kernel = Matern(para,"fixed")
            else:
                print("Unknown kernel !")
                print('Only the following kernel functions are legal')
                print('RBF | DotProduct | WhiteKernel | Matern')
        return kernel
    
    def Instantiate(self, ):
        """
        Helper function to instantiate the Gaussian Process Regressor and set the kernel parameters.
        
        Returns:
            An instantiated kernel function object.
        """
        Xtrain = np.array(self.target_data)
        Ytrain = np.array(self.target_response)
        ker = self.call_kernel()
        GPr = GPR(kernel=ker).fit(Xtrain,Ytrain)
        print('Trained Kernel Function :', GPr.kernel_)
        para = np.exp(GPr.kernel_.theta)

        Inst_kernel = self.call_kernel(para)
        return Inst_kernel


    def cal_beta(self, B=1,tao = None):
        """
        Helper function to calculate the beta values that balance the source and target data distributions.
        
        Args:
            B (float): Bound on the beta values. Defaults to 1.
        
        Returns:
            A numpy array containing the calculated beta values.
        """
        
        n = len(self.source_data)
        m = len(self.target_data)
        if tao is None:
            tao = B/np.sqrt(n)/2
        elif type(tao) == float:
            tao = tao
        else:
            print('The input parameter tao must be a float, like 0.1')
        
        """
        1-tao <= sum(beta_i) <= 1+tao

        E(beta) = 1 , E is expectation 
        """
        kernel = self.Instantiate()
        # n x n matrix
        K_SS = kernel(self.source_data)
        # m x n matrix
        K_ST = kernel(self.source_data,self.target_data)


        """
        standard quadratic programming problem 

        minimize    (1/2) * X.T * P * X + q.T * X
        subject to  G * X <= h
                    A * X = b

        M_P: a square matrix of size (n, n) representing the quadratic coefficients of the objective function. The matrix must be symmetric and positive semidefinite, meaning that x.T M_P x >= 0 for any vector x.

        V_q: a column vector of size (n, 1) representing the linear coefficients of the objective function.

        M_G: a matrix of size (m, n) representing the linear constraints of the problem. Each row of M_G represents a constraint, and each column represents a variable.

        h: a column vector of size (m, 1) representing the right-hand side of the linear constraints. Each element of h corresponds to a row of M_G.

        M_A: an optional matrix of size (p, n) representing additional linear equality constraints of the problem. Each row of M_A represents a constraint, and each column represents a variable.

        b: an optional column vector of size (p, 1) representing the right-hand side of the additional linear equality constraints. Each element of b corresponds to a row of M_A.
        """

        # In KMM 
        M_P =  K_SS
        # matrix multiplication
        V_q = - n/m * np.sum(K_ST,axis = 1)
        M_G1 = -np.eye(n)
        M_G2 = np.eye(n)
        M_A1 = np.ones((1,n))/n
        M_A2 = -np.ones((1,n))/n
        V_h1 = np.zeros((n,1))
        V_h2 = B * np.ones((n,1))

        # convert to matrix
        M_P = matrix(M_P)
        V_q = matrix(V_q)
        M_G1_U = copy.deepcopy(matrix(M_G1))
        M_G2_U = copy.deepcopy(matrix(M_G2))
        M_A1 = matrix(M_A1)
        M_A2 = matrix(M_A2)
        V_h1_U = copy.deepcopy(matrix(V_h1))
        V_h2_U = copy.deepcopy(matrix(V_h2))
        M_G = matrix(np.r_[M_G1_U, M_G2_U,M_A1,M_A2])
        h1 = matrix(1+tao)
        h2 = matrix(tao-1)
        h = matrix(np.r_[V_h1_U, V_h2_U,h1,h2])

        

        solvers.options['show_progress'] = True
        sol = solvers.qp(M_P,V_q,M_G,h)
        beta = sol['x']
        return np.array(beta)



from setuptools import setup, find_packages
setup(
    name='KMMTR',  # 包名
    version='0.0.1',  # 版本
    description="a transfer learning regression model based on Kernel Mean Matching (KMM) algorithm",  # 包简介
    long_description=open('README.md',encoding='utf-8').read(),  # 读取文件中介绍包的详细内容
    include_package_data=True,  # 是否允许上传资源文件
    author='CaoBin',  # 作者
    author_email='bcao@shu.edu.com',  # 作者邮件
    maintainer='CaoBin',  # 维护者
    maintainer_email='binjacobcao@gmail.com',  # 维护者邮件
    license='MIT License',  # 协议
    url='https://github.com/Bin-Cao/KMMTransferRegressor',  # github或者自己的网站地址
    packages=find_packages(include=['KMMTR', 'KMMTR.*']),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',  # 设置编写时的python版本
    ],
    python_requires='>=3.7',  # 设置python版本要求
    install_requires=['scikit-learn','pandas','numpy','cvxopt'],  # 安装所需要的库
    entry_points={
        'console_scripts': [
            ''],
    },  # 设置命令行工具(可不使用就可以注释掉)

)
